__version__ = "0.1.7"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(HEAD detached at v0.1.7)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '26fa564aea92a5640ab3cf1be333324566f41896'
__git_head_subject__ = 'Finished Release v0.1.7'
__git_head_message__ = ''
__git_dirty_diff__ = ''